/*
   Initial release:  10-MAR-2011
   By:               Jack Duijf
                     jack@jdsoftware.nl

   This DcRtfControl class facilitates a simple RTF editor in
   a Express++ application.
   It enables simple functions like:
      - fontselection
      - fontsize
      - bold
      - italic
      - Alignment
      - Bulletlist
      - Color selection

   Be aware that thic code is far from perfect......
   For me, this fits my needs.
   There are some issues that need to be addressed soem time in the furure.
      - Rulers
      - Resizing
      - Drag and drop
      - Open / save / save-as

   This code relies on Express++.
   Any express++ user is free to add this code to his/hers application at no costs.

   If you haven written additions to this code, pleas let me know.
   They might be usefull for other Express++ users.

 */

#include "Common.ch"
#include "DcDialog.ch"
#include "xbp.ch"
#include "DcRtfEdit.ch"

#define  BUTTONWIDTH    4.3
#define  RTFBUTTONSIZE  SIZE BUTTONWIDTH,1.43


CLASS DcRtfControl
HIDDEN:
VAR      oFontName                                                            // DCCOMBO
VAR      oFontSize                                                            // DCCOMBO
VAR      oBoldBtn                                                             // DCPUSHBUTTON
VAR      oItalicBtn                                                           // DCPUSHBUTTON
VAR      oUnderlineBtn                                                        // DCPUSHBUTTON
VAR      oAlignLeftBtn                                                        // DCPUSHBUTTON
VAR      oAlignCenterBtn                                                      // DCPUSHBUTTON
VAR      oAlignRightBtn                                                       // DCPUSHBUTTON
VAR      oBulletListBtn                                                       // DCPUSHBUTTON
VAR      oColorBtn                                                            // DCPUSHBUTTON
VAR      oColorActive                                                         // DCSTATIC
VAR      GetList
EXPORTED:
VAR      aRtfSize                                                             // Current DCRTF Size
VAR      aFontList                                                            // Available fonts         {"Arial","Verdana",...}
VAR      aFontSize                                                            // And sizes               {"8","9","10",...}
VAR      cFontName                                                            // Current RTF fontname    "Arial"
VAR      cFontSize                                                            // Current RTF fontsize    "9"
VAR      oRtf                                                                 // DCRTF                   Reference to xbpRtf

VAR      Row                                                                  // Top-left row of control
VAR      Col                                                                  // Top-left column of control
VAR      Width                                                                // Width of control
VAr      Depth                                                                // Depth of control
VAR      Caption                                                              // Caption of control (Groupbox)
VAR      Font                                                                 // Font  of control (Groupbox)
VAR      FgColor                                                              // Color of control (Groupbox)
VAR      BgColor                                                              // Color of control (Groupbox)
VAR      Group                                                                // Group for DCREAD
VAR      Id                                                                   // Id    fot DCREAD
VAR      BulletIndent                                                         // Bullet indent in pixel (Default 25)
VAR      HangingIndent                                                        // HanginIndent in pixel (Default 25)
VAR      MaxLength                                                            // Maximum length of buffer
VAR      RightMargin                                                          // Maximum length of buffer
VAR      EnableAlign                                                          // Enable align controls
VAR      EnableBullet                                                         // Enable bullet control
VAR      EnableColor                                                          // Enable color control (future use)
VAR      EnableFont                                                           // Enable font control
VAR      WhenBlock                                                            // Code-block for WHEN clause
VAR      HideBlock                                                            // Code-block for HIDE clause
VAR      EditProtectBlock                                                     // Code-block for Edit-protect clause
METHOD   Init
METHOD   Destroy
METHOD   BuildGets                                                            // Build getlist
METHOD   SetDefaults                                                          // Setup control defaults
METHOD   RtfDefaults                                                          // RTF defaults
METHOD   RtfUpdate                                                            // Refresh buttons status
METHOD   SetParameters                                                        // All settings in 1 call
METHOD   ColorPicker                                                          // Select foreground color
ENDCLASS

METHOD DcRtfControl:Init(GetList)
LOCAL aFonts         := {}
::GetList            := GetList                                               // Active getlist
::Row                := 0
::Col                := 0
::WhenBlock          := {||TRUE}                                              // Code-block for WHEN clause (all sub-controls)
::HideBlock          := {||FALSE}                                             // Code-block for HIDE clause (for main static)
::EditProtectBlock   := {||FALSE}                                             // Code-block for Edit-protect clause (DCRTF)
::aFontList          := {}
aFonts               := XbpFont():New():List()                                // Get available fonts
Aeval(aFonts,{|o|if(Ascan(::aFontList,o:FamilyName)==0,AAdd(::aFontList,o:FamilyName), NIL) } )
ASort(::aFontList)                                                            // Sort by name
::aFontSize          := { "6", "8","10","11","12","14","16","18",;            // Available font sizes
                          "20","22","24","32","36"}
::SetDefaults()                                                               // General default for the control
Return SELF

METHOD DcRtfControl:Destroy()
::GetList            := nil
Asize(::aFontList,0)
Asize(::aFontSize,0)
::oFontName          := nil
::oFontSize          := nil
::oBoldBtn           := nil
::oItalicBtn         := nil
::oUnderlineBtn      := nil
::oAlignLeftBtn      := nil
::oAlignCenterBtn    := nil
::oAlignRightBtn     := nil
::oBulletListBtn     := nil
::oColorBtn          := nil
::oColorActive       := nil
::oRtf               := nil
::Font               := ""
Return

METHOD DcRtfControl:SetDefaults()
::cFontName          := "Arial"                                               // RTF fontname
::cFontSize          := "8"                                                   // RTF fontsize
::EnableAlign        := TRUE                                                  // Enable align controls
::EnableBullet       := TRUE                                                  // Enable bullet controls
::EnableColor        := TRUE                                                  // Enable color controls  (future use)
::EnableFont         := TRUE                                                  // Enable font controls
::Width              := 80                                                    // Assume character based, not PIXEL
::Depth              := 12
::Caption            := ""                                                    // No box Caption
::FgColor            := GRA_CLR_BLACK
::BgColor            := XBPSYSCLR_TRANSPARENT
::Id                 := nil
::Group              := nil
::BulletIndent       := 25
::HangingIndent      := 25
::MaxLength          := 0                                                     // Maximum length of buffer (unlimited)
::RightMargin        := 0                                                     // No right margin

Return

METHOD DcRtfControl:RtfDefaults(oRtf)                                         // xbpRtf defaults
oRtf:BulletIndent    := ::BulletIndent                                        // In dots
oRtf:scrollbars      := XBP_SCROLLBAR_HORIZ + XBP_SCROLLBAR_VERT              // Auto scrollbars
oRtf:SelHangingIndent:= ::HangingIndent
oRtf:MaxLength       := ::MaxLength
oRtf:RightMargin     := ::RightMargin
Return
//
// The method :SetParameters() exists only to facilitate the #command for DCRTFEDIT
//
METHOD DcRtfControl:SetParameters(nWidth,nDepth,cCaption,uFont,bWhen,bHide,nFgColor,nBgColor,lAlign,lBullet,lFont,lColor,cId,cGroup,nBulletInd,nHangingInd)
Default nWidth       To ::Width
Default nDepth       To ::Depth
Default cCaption     To ::Caption
Default uFont        To ::Font
Default bWhen        To ::WhenBlock
Default bHide        To ::HideBlock
Default nFgColor     To ::FgColor
Default nBgColor     To ::BgColor
Default lAlign       To !::EnableAlign
Default lBullet      To !::EnableBullet
Default lFont        To !::EnableFont
Default lColor       To !::EnableColor
Default cId          To ::Id
Default cGroup       To ::Group
Default nBulletInd   To ::BulletIndent
Default nHangingInd  To ::HangingIndent
::Width              := nWidth
::Depth              := nDepth
::Caption            := cCaption
::Font               := uFont
::WhenBlock          := bWhen
::HideBlock          := bHide
::FgColor            := nFgColor
::BgColor            := nBgColor
::EnableAlign        := !lAlign
::EnableBullet       := !lBullet
::EnableFont         := !lFont
::EnableColor        := !lColor
::Id                 := cId
::Group              := cGroup
::BulletIndent       := nBulletInd
::HangingIndent      := nHangingInd
Return

METHOD DcRtfControl:BuildGets(cData,oParent)
LOCAL nLine          := 2.5                                                   // Start of DCRTF
LOCAL nCol           := 1
LOCAL oStaticBox     := nil
LOCAL GetList        := ::GetList
LOCAL bRefresh       := {||::RtfUpdate(),SetAppFocus(::oRtf)}                 // Refresh Controls
LOCAL bBitmapOnOff   := {|l,a,b|iif(iif(ValType(l) = "L",l,FALSE),a,b)}
LOCAL bBitmapOption  := {|n,o,a,b|iif(iif(ValType(n) = "N",(n=o),FALSE),a,b)}
LOCAL bToggle        := {|l|iif(ValType(l)="L",!l,iif(Empty(l),TRUE,FALSE))}  // Toggle on/off setting
LOCAL bSetAlignment  := {|n|::oRtf:SelAlignment := n,Eval(bRefresh)}          // Set Left/Right/Center alingment
LOCAL oColorBtnFg    := nil
LOCAL bHideToolbar   := {||!Eval(::WhenBlock)}
LOCAL aSize          := {}
LOCAL bSizeRtf       := {||aSize := Aclone(::aRtfSize), aSize[2] += iif(Eval(bHideToolbar),30,0),::oRtf:SetSize(aSize),::oRtf:Locked := Eval(bHideToolbar)}

@ ::Row  ,::Col      DCGROUP oStaticBox CAPTION ::Caption SIZE ::Width,::Depth PARENT oParent HIDE ::HideBlock ;
                     FONT ::Font COLOR ::FgColor,::BgColor ID ::Id GROUP ::Group

@ nLine  ,nCol       DCRTF VAR cData                     ;
                     SIZE ::Width-2,::Depth-nLine-.25    ;
                     OBJECT ::oRtf                       ;
                     PARENT oStaticBox                   ;
                     SELCHANGE {|a,b,oRtf|::RtfUpdate()} ;
                     EVAL {|o|::RtfDefaults(o),::aRtfSize := o:CurrentSize()}

nLine                := 1
if ::EnableFont                                                               // Buttons for change font and fontsize
   @ nLine,nCol      DCCOMBOBOX ::cFontName LIST ::aFontList FONT '9.Arial' ;
                     SIZE 20,10                          ;
                     PARENT oStaticBox                   ;
                     ID 'FONTNAME'                       ;
                     ITEMSELECTED {|a,b,o|::oRTF:SelFontName := Trim(o:XbpSLE:GetData())} ;
                     OBJECT ::oFontName                  ;
                     HIDE bHideToolbar

   nCol              += 20
   @ nLine,nCol      DCCOMBOBOX ::cFontSize LIST ::aFontSize FONT '9.Arial' ;
                     SIZE  6,10                          ;
                     PARENT oStaticBox                   ;
                     ID 'FONTSIZE'                       ;
                     ITEMSELECTED {|a,b,o|::oRtf:SelFontSize := Val(o:XbpSLE:GetData()+".00")} ;
                     OBJECT ::oFontSize                  ;
                     HIDE bHideToolbar
   nCol              += 7
endif
nLine                -= 0.2
@ nLine,nCol         DCPUSHBUTTON CAPTION {||Eval(bBitmapOnOff,::oRTF:SelBold,BITMAP_RTF_BOLD_ACTIVE,BITMAP_RTF_BOLD_INACTIVE)} ;
                                  OBJECT ::oBoldBtn      ;
                                  RTFBUTTONSIZE          ;
                                  ACTION {||::oRTF:SelBold  := Eval(bToggle,::oRTF:SelBold),Eval(bRefresh)} ;
                                  PARENT oStaticBox      ;
                                  SCALEBITMAP            ;
                                  HIDE bHideToolbar
nCol                 += BUTTONWIDTH
@ nLine,nCol         DCPUSHBUTTON CAPTION {||Eval(bBitmapOnOff,::oRtf:SelItalic,BITMAP_RTF_ITALIC_ACTIVE,BITMAP_RTF_ITALIC_INACTIVE)} ;
                                  OBJECT ::oItalicBtn    ;
                                  RTFBUTTONSIZE          ;
                                  ACTION {||::oRTF:SelItalic := Eval(bToggle,::oRTF:SelItalic),Eval(bRefresh)} ;
                                  PARENT oStaticBox      ;
                                  SCALEBITMAP            ;
                                  HIDE bHideToolbar
nCol                 += BUTTONWIDTH
@ nLine,nCol         DCPUSHBUTTON CAPTION {||Eval(bBitmapOnOff,::oRtf:SelUnderline,BITMAP_RTF_UNDERLINE_ACTIVE,BITMAP_RTF_UNDERLINE_INACTIVE)};
                                  OBJECT ::oUnderlineBtn ;
                                  RTFBUTTONSIZE          ;
                                  ACTION {||::oRTF:SelUnderline := Eval(bToggle,::oRTF:SelUnderline),Eval(bRefresh)};
                                  PARENT oStaticBox      ;
                                  SCALEBITMAP            ;
                                  HIDE bHideToolbar
nCol                 += BUTTONWIDTH
if ::EnableAlign                                                              // Buttons for change allignment
   @ nLine,nCol      DCPUSHBUTTON CAPTION {||Eval(bBitmapOption,::oRtf:SelAlignment,XBPRTF_ALIGN_LEFT,BITMAP_RTF_ALIGNLEFT_ACTIVE,BITMAP_RTF_ALIGNLEFT_INACTIVE)} ;
                                  OBJECT ::oAlignLeftBtn ;
                                  RTFBUTTONSIZE          ;
                                  ACTION {||Eval(bSetAlignment,XBPRTF_ALIGN_LEFT)} ;
                                  PARENT oStaticBox      ;
                                  SCALEBITMAP            ;
                                  HIDE bHideToolbar
   nCol              += BUTTONWIDTH
   @ nLine,nCol      DCPUSHBUTTON CAPTION {||Eval(bBitmapOption,::oRtf:SelAlignment,XBPRTF_ALIGN_CENTER,BITMAP_RTF_ALIGNCENTER_ACTIVE,BITMAP_RTF_ALIGNCENTER_INACTIVE)};
                                  OBJECT ::oAlignCenterBtn ;
                                  RTFBUTTONSIZE          ;
                                  ACTION {||Eval(bSetAlignment,XBPRTF_ALIGN_CENTER)} ;
                                  PARENT oStaticBox      ;
                                  SCALEBITMAP            ;
                                  HIDE bHideToolbar
   nCol              += BUTTONWIDTH
   @ nLine,nCol      DCPUSHBUTTON CAPTION {||Eval(bBitmapOption,::oRtf:SelAlignment,XBPRTF_ALIGN_RIGHT,BITMAP_RTF_ALIGNRIGHT_ACTIVE,BITMAP_RTF_ALIGNRIGHT_INACTIVE)};
                                  OBJECT ::oAlignRightBtn ;
                                  RTFBUTTONSIZE          ;
                                  ACTION {||Eval(bSetAlignment,XBPRTF_ALIGN_RIGHT)} ;
                                  PARENT oStaticBox      ;
                                  SCALEBITMAP            ;
                                  HIDE bHideToolbar
   nCol              += BUTTONWIDTH
endif
if ::EnableBullet                                                             // Buttons for bulletlist
   @ nLine,nCol      DCPUSHBUTTON CAPTION {||Eval(bBitmapOnOff,::oRtf:SelBullet,BITMAP_RTF_BULLITLIST_ACTIVE,BITMAP_RTF_BULLITLIST_INACTIVE)};
                                  OBJECT ::oBulletListBtn ;
                                  RTFBUTTONSIZE          ;
                                  ACTION {||::oRTF:SelBullet := Eval(bToggle,::oRTF:SelBullet),Eval(bRefresh)} ;
                                  PARENT oStaticBox      ;
                                  SCALEBITMAP            ;
                                  HIDE bHideToolbar
   nCol              += BUTTONWIDTH
endif
if ::EnableColor
   @ nLine+1.15,nCol      DCSTATIC    TYPE XBPSTATIC_TYPE_FGNDRECT ;
                                 SIZE BUTTONWIDTH,0.3    ;
                                 OBJECT ::oColorActive   ;
                                 PARENT oStaticBox       ;
                                 COLOR {||{::FgColor,::FgColor}} ;
                                 HIDE {||Eval(bHideToolbar)}

   @ nLine,nCol      DCPUSHBUTTON CAPTION "C"            ;
                                  OBJECT ::oColorBtn     ;
                                  SIZE BUTTONWIDTH,1.2   ;
                                  PARENT oStaticBox      ;
                                  ACTION {||::ColorPicker()} OBJECT oColorBtnFg ;
                                  SCALEBITMAP            ;
                                  HIDE {||Eval(bHideToolbar)}
   nCol              += BUTTONWIDTH
endif
//
// This dummy control is responsible for updating the controll while moving arround the RTF document
//
@ nLine,nCol         DCPUSHBUTTON CAPTION "Dummy"        ;
                                 RTFBUTTONSIZE           ;
                                 HIDE {||Eval(bSizeRtf),TRUE} ;
                                 EVAL {||::RtfUpdate()}
nCol                 += BUTTONWIDTH
Return

METHOD DcRtfControl:ColorPicker()
LOCAL aColor         := Dc_PopColor(,2,TRUE)
If ValType(aColor) = "A" .and. Len(aColor) = 3
   ::FgColor            := GraMakeRGBColor(aColor)
   ::oRtf:SelColor      := ::FgColor
   Dc_GetRefresh(::oColorActive)
Endif
SetAppFocus(::oRtf)
Return

METHOD DcRtfControl:RtfUpdate()
LOCAL cTmp           := ""

if ValType(::oFontName) = "O"
   cTmp              := ::oRtf:SelFontName                                    // Current selected font (in document)
   if Empty(cTmp)                                                             // Multiple fonts selected
      cTmp           := ""                                                    // Clear fontname
   endif
   if !(::cFontname == cTmp)                                                  // Different font
      ::cFontName    := cTmp                                                  // Change fontname
      Dc_GetRefresh(::oFontName)                                              // And redisplay
   endif
endif
if ValType(::oFontSize) = "O"
   cTmp              := ::oRtf:SelFontSize                                    // Current selected fontsize (in document)
   if Empty(cTmp)                                                             // Multiple fontsizes selected
      cTmp           := ""                                                    // Clear fontsize
   else                                                                       // Font in 2 decimals
      cTmp           := AllTrim(Str(cTmp,3,0))                                // To string with no decimals
   endif
   if !(::cFontSize == cTmp)                                                  // Different fontsize
      ::cFontSize    := cTmp                                                  // Change fontsize
      Dc_GetRefresh(::oFontSize)                                              // And redisplay
   endif
endif
::FgColor         := ::oRtf:SelColor
Dc_GetRefresh(::oBoldBtn)                                                     // Refresh button status
Dc_GetRefresh(::oItalicBtn)
Dc_GetRefresh(::oUnderlineBtn)
Dc_GetRefresh(::oAlignLeftBtn)
Dc_GetRefresh(::oAlignCenterBtn)
Dc_GetRefresh(::oAlignRightBtn)
Dc_GetRefresh(::oBulletListBtn)
Dc_GetRefresh(::oColorBtn)
Dc_GetRefresh(::oColorActive)
RETURN Nil

